﻿namespace DataModel
{
  using System;
  using System.Collections.Generic;
  using System.ComponentModel.DataAnnotations;
  using System.Data;
  using System.Data.Objects;
  using System.Data.SqlClient;
  using System.Linq;
  using System.Text.RegularExpressions;
  using Unleashed.EntityFramework;

  public partial class NorthwindEntities : UnleashedObjectContext
  {
    private static readonly Dictionary<string, ValidationResult> ConstraintErrors =
      new Dictionary<string, ValidationResult>
        {
        { "UK_Products",               new ValidationResult("Product Name must be unique.", new string[] { "ProductName" }) },
        { "FK_Order_Details_Products", new ValidationResult("Product cannot be deleted because it is referenced in one or more Orders.") },
        { "FK_Products_Suppliers",     new ValidationResult("Supplier cannot be deleted because it is referenced in one or more Products.") }
      };

/*  
    // Original version of the FulfillOrder method implemented in .NET code
    public void FulfillOrder(int orderId)
    {
      Order order = this.Orders
        .Include("Order_Details").Include("Order_Details.Product")
        .First(o => o.OrderID == orderId);

      this.Validate(order);

      if (order.OrderStatus != (byte)DataModel.OrderStatus.Paid)
      {
        throw new ValidationException(
          new ValidationResult("Order status must be Paid for the order to be fulfilled", new[] { "OrderStatus" }),
          null, order.OrderStatus);
      }

      order.OrderStatus = (byte)DataModel.OrderStatus.Fulfilled;
      order.ShippedDate = DateTime.Now;
      foreach (Order_Detail orderItem in order.Order_Details)
        orderItem.Product.UnitsInStock -= orderItem.Quantity;
    }
*/

    // Optimized version of the FulfillOrder method, reimplemented in T-SQL
    public void FulfillOrder(int orderId)
    {
      try
      {
        this.SqlFulfillOrder(orderId);
      }
      catch (EntityCommandExecutionException e)
      {
        HandleSqlException(e);
        throw;
      }
    }

    public void FulfillOrder(Order order)
    {
#if DEBUG
      this.Validate(order);
#endif

      if (order.OrderStatus != (byte)DataModel.OrderStatus.Paid)
      {
        throw new InvalidOperationException(
          "Order status must be Paid for the order to be fulfilled");
      }

      order.OrderStatus = (byte)DataModel.OrderStatus.Fulfilled;
      order.ShippedDate = DateTime.Now;
      IQueryable<Order_Detail> orderItems = this.Order_Details
        .Where(od => od.OrderID == order.OrderID);
      foreach (Order_Detail orderItem in orderItems)
        orderItem.Product.UnitsInStock -= orderItem.Quantity;
    }
    
    public override int SaveChanges(SaveOptions options)
    {
      try
      {
        return base.SaveChanges(options);
      }
      catch (UpdateException e)
      {
        HandleSqlException(e);
        throw;
      }
    }

    private static void HandleSqlException(Exception e)
    {
      var error = e.InnerException as SqlException;
      if (error != null)
      {
        if (error.Number == 547)
          HandleReferenceConstraintViolation(error);
        else if (error.Number == 2627)
          HandleUniqueKeyConstraintViolation(error);
        else if (error.Number >= 50000 && error.Class <= 16)
          throw new ValidationException(error.Message);      
      }      
    }

    // The DELETE statement conflicted with the REFERENCE constraint "FK_Products_Suppliers". The conflict occurred in database "Northwind", table "dbo.Products", column 'SupplierID'.
    private static void HandleReferenceConstraintViolation(SqlException error)
    {
      Match match = Regex.Match(error.Message, "The .+? statement conflicted with the REFERENCE constraint \"(?<Constraint>.+?)\"", RegexOptions.IgnoreCase);
      string constraint = match.Groups["Constraint"].Value;
      ValidationResult validationResult;
      if (ConstraintErrors.TryGetValue(constraint, out validationResult))
        ThrowValidationException(null, new[] { validationResult });
    }

    // Violation of UNIQUE KEY constraint 'UK_Products'. Cannot insert duplicate key in object 'dbo.Products'.
    private static void HandleUniqueKeyConstraintViolation(SqlException error)
    {
      Match match = Regex.Match(error.Message, "Violation of UNIQUE KEY constraint '(?<Constraint>.+?)'", RegexOptions.IgnoreCase);
      string constraint = match.Groups["Constraint"].Value;
      ValidationResult validationResult;
      if (ConstraintErrors.TryGetValue(constraint, out validationResult))
        ThrowValidationException(null, new [] { validationResult });
    }
  }
}
