﻿namespace Unleashed.DataAnnotations.Tests
{
  using System.Linq;
  using System.Security.Principal;
  using Microsoft.VisualStudio.TestTools.UnitTesting;

  [TestClass]
  public class AuthorizationAttributeTest
  {
    [Authorization(Deny.Users, "Adam")]
    private class ClassWithAllOperationsDeniedToAdam { }

    [TestMethod]
    public void IsAuthorizedReturnsTrueByDefault()
    {
      var adam = new GenericPrincipal(new GenericIdentity("Udi"), null);
      var attribute = typeof(ClassWithAllOperationsDeniedToAdam).GetCustomAttributes(false).OfType<AuthorizationAttribute>().Single();
      Assert.IsTrue(attribute.IsAuthorized(adam));
    }

    [TestMethod]
    public void IsAuthorizedReturnsFalseForMatchingDeny()
    {
      var adam = new GenericPrincipal(new GenericIdentity("Adam"), null);
      var attribute = typeof(ClassWithAllOperationsDeniedToAdam).GetCustomAttributes(false).OfType<AuthorizationAttribute>().Single();
      Assert.IsFalse(attribute.IsAuthorized(adam));
    }

    [Authorization(Allow.Users, "Udi", Deny.Users, "Adam")]
    private class ClassWithMultipleRules { }

    [TestMethod]
    public void IsAuthorizedSupportsMultipleRules()
    {
      var adam = new GenericPrincipal(new GenericIdentity("Adam"), null);
      var udi = new GenericPrincipal(new GenericIdentity("Udi"), null);
      var attribute = typeof(ClassWithMultipleRules).GetCustomAttributes(false).OfType<AuthorizationAttribute>().Single();
      Assert.IsFalse(attribute.IsAuthorized(adam));
      Assert.IsTrue(attribute.IsAuthorized(udi));
    }

    [Authorization(Allow.Users, Actions.Read, "Udi", Deny.Users, "*")]
    class TestClassWithReadAllowedToUdi { }

    [TestMethod]
    public void IsAuthorizedReturnsFalseWhenOperationDoesntMatch()
    {
      var udi = new GenericPrincipal(new GenericIdentity("Udi"), null);
      var attribute = typeof(TestClassWithReadAllowedToUdi).GetCustomAttributes(false).OfType<AuthorizationAttribute>().Single();
      Assert.IsTrue(attribute.IsAuthorized(udi, Actions.Read));
      Assert.IsFalse(attribute.IsAuthorized(udi, Actions.Update));
    }

    [Authorization(Allow.Users, "Udi", Deny.Users, "*")]
    class TestClassWithAllOperationsAllowedToUdi { }

    [TestMethod]
    public void IsAuthorizedReturnsTrueWhenOperationMatchesSet()
    {
      var udi = new GenericPrincipal(new GenericIdentity("Udi"), null);
      var adam = new GenericPrincipal(new GenericIdentity("Adam"), null);
      var attribute = typeof(TestClassWithAllOperationsAllowedToUdi).GetCustomAttributes(false).OfType<AuthorizationAttribute>().Single();
      Assert.IsTrue(attribute.IsAuthorized(udi, Actions.Update));
      Assert.IsFalse(attribute.IsAuthorized(adam, Actions.Update));
    }

    [Authorization(Allow.Users, Actions.Read, "Udi", Allow.Users, Actions.Update, "Udi", Deny.Users, "*")]
    class ClassWithMultiplePartialRules { }

    [TestMethod]
    public void IsAuthorizedReturnsTrueWhenOperationsPartiallyMatchMultipleRules()
    {
      var udi = new GenericPrincipal(new GenericIdentity("Udi"), null);
      var attribute = typeof(ClassWithMultiplePartialRules).GetCustomAttributes(false).OfType<AuthorizationAttribute>().Single();
      Assert.IsTrue(attribute.IsAuthorized(udi, Actions.Read | Actions.Update));
    }

    [Authorization(Deny.Users, Actions.Update, "Udi")]
    class TestClassWithUpdateDeniedToUdi { }

    [TestMethod]
    public void IsAuthorizedReturnsFalseWhenOperationIsDenied()
    {
      var udi = new GenericPrincipal(new GenericIdentity("Udi"), null);
      var attribute = typeof(TestClassWithUpdateDeniedToUdi).GetCustomAttributes(false).OfType<AuthorizationAttribute>().Single();
      Assert.IsTrue(attribute.IsAuthorized(udi, Actions.Read));
      Assert.IsFalse(attribute.IsAuthorized(udi, Actions.Update));
    }
  }
}