﻿using System.Text;
using System.Collections.Generic;
using System.Linq;
using Microsoft.VisualStudio.TestTools.UnitTesting;

namespace Unleashed.DataAnnotations.Tests
{
  using System;

  [TestClass]
  public class CustomQueryAttributeTest
  {
    private class FakeClass
    {
      public static IQueryable ExpectedResult;
      public static IQueryable ActualSource;

      public static IQueryable GetQuery(IQueryable source)
      {
        ActualSource = source;
        return ExpectedResult;
      }

      public static void GetQueryWithoutResult(IQueryable source)
      {
        ActualSource = source;
      }

      public static IQueryable GetQueryWithoutParameters()
      {
        return ExpectedResult;
      }
    }

    [TestMethod]
    public void ConstructorInitializesProperties()
    {
      var attribute = new CustomQueryAttribute(typeof(FakeClass), "GetQuery");
      Assert.AreEqual(typeof(FakeClass), attribute.Type);
      Assert.AreEqual("GetQuery", attribute.Method);
    }

    [TestMethod]
    public void GetQueryCallsTheQueryMethod()
    {
      IQueryable originalQuery = new List<string>().AsQueryable();
      FakeClass.ExpectedResult = new List<string>().AsQueryable();
      var attribute = new CustomQueryAttribute(typeof(FakeClass), "GetQuery");
      Assert.AreEqual(FakeClass.ExpectedResult, attribute.GetQueryable(originalQuery));
      Assert.AreEqual(originalQuery, FakeClass.ActualSource);
    }

    [TestMethod]
    public void GetQueryThrowsExceptionWhenTypeNotSpecified()
    {
      var attribute = new CustomQueryAttribute(null, "GetQuery");
      try
      {
        attribute.GetQueryable(new List<string>().AsQueryable());
        Assert.Fail();
      }
      catch (Exception e)
      {
        Assert.AreEqual(typeof(InvalidOperationException), e.GetType());
        StringAssert.Contains(e.Message, "CustomQueryAttribute.Type");
      }
    }

    [TestMethod]
    public void GetQueryThrowsExceptionWhenMethodNotSpecified()
    {
      var attribute = new CustomQueryAttribute(typeof(FakeClass), null);
      try
      {
        attribute.GetQueryable(new List<string>().AsQueryable());
        Assert.Fail();
      }
      catch (Exception e)
      {
        Assert.AreEqual(typeof(InvalidOperationException), e.GetType());
        StringAssert.Contains(e.Message, "CustomQueryAttribute.Method");
      }      
    }

    [TestMethod]
    public void GetQueryThrowsExceptionWhenMethodCannotBeFound()
    {
      var attribute = new CustomQueryAttribute(typeof(FakeClass), "NonExistentMethod");
      try
      {
        attribute.GetQueryable(new List<string>().AsQueryable());
        Assert.Fail();
      }
      catch (Exception e)
      {
        Assert.AreEqual(typeof(InvalidOperationException), e.GetType());
        StringAssert.Contains(e.Message, "does not exist");
        StringAssert.Contains(e.Message, "not public and static");
      }            
    }

    [TestMethod]
    public void GetQueryThrowsExceptionWhenMethodDoesNotReturnIQueryable()
    {
      var attribute = new CustomQueryAttribute(typeof(FakeClass), "GetQueryWithoutResult");
      try
      {
        attribute.GetQueryable(new List<string>().AsQueryable());
        Assert.Fail();
      }
      catch (Exception e)
      {
        Assert.AreEqual(typeof(InvalidOperationException), e.GetType());
        StringAssert.Contains(e.Message, "return");
        StringAssert.Contains(e.Message, "IQueryable");
      }                  
    }

    [TestMethod]
    public void GetQueryThrowsExceptionWhenMethodDoesNotHaveParameter()
    {
      var attribute = new CustomQueryAttribute(typeof(FakeClass), "GetQueryWithoutParameters");
      try
      {
        attribute.GetQueryable(new List<string>().AsQueryable());
        Assert.Fail();
      }
      catch (Exception e)
      {
        Assert.AreEqual(typeof(InvalidOperationException), e.GetType());
        StringAssert.Contains(e.Message, "method");
        StringAssert.Contains(e.Message, "expected signature");
      }                        
    }
  }
}
