﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Data.Objects;
using System.Data;
using System.Transactions;
using System.Threading.Tasks;

namespace EntityFrameworkSamples
{
  class Program
  {
    static void SimpleQuery()
    {
      using (NorthwindEntities context = new NorthwindEntities())
      {
        var query = from p in context.Products
                    where p.UnitsInStock < p.ReorderLevel
                    select p;

        foreach (Product p in query)
          Console.WriteLine(p.ProductName);
      }
    }

    static void QueryAccrossRelationships()
    {
      using (NorthwindEntities context = new NorthwindEntities())
      {
        var query = from p in context.Products
                    from s in context.Suppliers
                    where p.SupplierID == s.SupplierID 
                       && s.City == "London"
                    select p;

        foreach (Product p in query)
          Console.WriteLine(p.ProductName);
      }


      using (NorthwindEntities context = new NorthwindEntities())
      {
        var query = from p in context.Products
                    where p.Supplier.City == "London"
                    select p;

        foreach (Product p in query)
          Console.WriteLine(p.ProductName);
      }
    }

    static void ModifyingAndSubmittingChanges()
    {
      using (NorthwindEntities context = new NorthwindEntities())
      {
        var query = from p in context.Products
                    where p.ProductName == "Outback Lager"
                    select p;

        Product outbackLager = query.First();

        outbackLager.UnitsInStock += outbackLager.UnitsOnOrder;

        context.SaveChanges();
      }

      using (NorthwindEntities context = new NorthwindEntities())
      {        
        Product p = new Product() { ProductName = "Mukuzani" };
        context.Products.AddObject(p);
        context.SaveChanges();
      }

      using (NorthwindEntities context = new NorthwindEntities())
      {
        var query = from p in context.Products
                    where p.ProductName == "Mukuzani"
                    select p;

        Product product = query.First();
        context.Products.DeleteObject(product);
        context.SaveChanges();
      }
    }

    static void QueryExecution()
    {
      using (NorthwindEntities context = new NorthwindEntities())
      {
        IQueryable<Product> query = from p in context.Products
                                    where p.Supplier.City == "London"
                                    select p;

        Console.WriteLine(query.Expression);

        foreach (Product p in query)
          Console.WriteLine(p.ProductName);

        Product product = query.First();
        Console.WriteLine(product.ProductName);

        List<Product> products = query.ToList();
        foreach (Product p in products)
          Console.WriteLine(p.ProductName);
      }
    }

    static void ObjectIdentity()
    {
      using (NorthwindEntities context = new NorthwindEntities())
      {
        Product chai1 = (from p in context.Products
                        where p.ProductName == "Chai"
                        select p).First();

        Product chai2 = (from p in context.Products
                         where p.ProductID == 1
                         select p).First();

        if (object.ReferenceEquals(chai1, chai2))
          Console.WriteLine("chai1 and chai2 are the same object");

        Product chai = (from p in context.Products
                        where p.ProductName == "Chai"
                        select p).First();

        chai.UnitsInStock += 50;


        List<Product> allProducts = (from p in context.Products
                                    select p).ToList();

        Console.WriteLine(chai.UnitsInStock);
      }      
    }

    static void DisableChangeTracking()
    {
      using (NorthwindEntities context = new NorthwindEntities())
      {
        context.Products.MergeOption = MergeOption.NoTracking;

        var chaiQuery = from p in context.Products
                        where p.ProductName == "Chai"
                        select p;

        Product chai1 = chaiQuery.First();
        Product chai2 = chaiQuery.First();
        if (!object.ReferenceEquals(chai1, chai2))
          Console.WriteLine("chai1 and chai2 are different objects");
      }
    }

    static void LazyLoading()
    {
      using (NorthwindEntities context = new NorthwindEntities())
      {
        var query = from p in context.Products
                    select p;

        foreach (Product p in query)
          Console.WriteLine(p.ProductName + " " + p.Supplier.CompanyName);
      }
    }

    static void EagerLoading()
    {
      using (NorthwindEntities context = new NorthwindEntities())
      {
        var query = from p in context.Products.Include("Supplier")
                    select p;        

        foreach (Product p in query)
          Console.WriteLine(p.ProductName + " " + p.Supplier.CompanyName);
      }
    }

    static void Projections()
    {
      using (NorthwindEntities context = new NorthwindEntities())
      {
        var query = from s in context.Suppliers
                    select new { s.CompanyName, s.City };

        foreach (var item in query)
          Console.WriteLine(item.CompanyName + " " + item.City );
      }
    }

    class SupplierProjection
    {
      public string CompanyName = null; 
      public string City = null;
    }

    static void ProjectionsExplicitClass()
    {
      using (NorthwindEntities context = new NorthwindEntities())
      {
        var query = from s in context.Suppliers
                    select new SupplierProjection()
                    { 
                      CompanyName = s.CompanyName, 
                      City = s.City 
                    };

        foreach (var item in query)
          Console.WriteLine(item.CompanyName + " " + item.City);
      }
    }

    static void Sorting()
    {
      using (NorthwindEntities context = new NorthwindEntities())
      {
        var query = from p in context.Products
                    orderby p.UnitPrice descending
                    select p; 
        foreach (Product p in query)
          Console.WriteLine(p.ProductName + " " + p.UnitPrice);
      }

      using (NorthwindEntities context = new NorthwindEntities())
      {
        var query = from p in context.Products
                    orderby p.UnitsInStock, p.UnitsOnOrder descending
                    select p;
        foreach (Product p in query)
          Console.WriteLine(p.ProductName + " " + p.UnitsInStock);
      }
    }

    static void Grouping()
    {      
      using (NorthwindEntities context = new NorthwindEntities())
      {
        var query = from p in context.Products
                    group p by p.SupplierID into g
                    select new { Supplier = g.FirstOrDefault().Supplier, Products = g };
        foreach (var group in query)
        {
          Console.WriteLine("Supplier: " + group.Supplier.CompanyName);
          foreach (var product in group.Products)
            Console.WriteLine("\t" + product.ProductName);
        }
      }
    }

    static void Joins()
    {
      using (NorthwindEntities context = new NorthwindEntities())
      {
        var query = from c in context.Customers
                    join s in context.Suppliers on c.City equals s.City
                    select new { 
                      c.City, 
                      Customer = c.CompanyName, 
                      Supplier = s.CompanyName
                    };
        foreach (var item in query)
          Console.WriteLine(item.City + " " + item.Customer + " " + item.Supplier);
      }

      using (NorthwindEntities context = new NorthwindEntities())
      {
        var query = from c in context.Customers
                    from s in context.Suppliers 
                    where c.City == s.City
                    select new {
                      c.City,
                      Customer = c.CompanyName,
                      Supplier = s.CompanyName
                    };

        foreach (var item in query)
          Console.WriteLine(item.City + " " + item.Customer + " " + item.Supplier);
      }
    }

    static void CompositeQueries()
    {
      using (NorthwindEntities context = new NorthwindEntities())
      {
        string city = "London";
        bool sortByContactName = true;

        var query = from c in context.Customers
                    select c;

        if (city != null)
        {
          query = from c in query
                  where c.City == city
                  select c;
        }

        if (sortByContactName)
        {
          query = from c in query
                  orderby c.ContactName
                  select c;
        }
        else
        {
          query = from c in query
                  orderby c.CompanyName
                  select c;
        }

        foreach(Customer c in query)
          Console.WriteLine(c.CompanyName + " " + c.ContactName);
      }
    }

    static void SqlTranslation()
    {
      using (NorthwindEntities context = new NorthwindEntities())
      {
        var query = from c in context.Customers
                    where c.City == "London"
                    orderby c.CompanyName
                    select c;

        ObjectQuery objectQuery = (ObjectQuery)query;
        Console.WriteLine(objectQuery.ToTraceString());
      }

      using (NorthwindEntities context = new NorthwindEntities())
      {
        string city = "London";
        var query = from c in context.Customers
                    where c.City == city
                    select c;

        ObjectQuery objectQuery = (ObjectQuery)query;
        Console.WriteLine(objectQuery.ToTraceString());
      }

      {
        string city = "'; drop database Northwind; select '";
        string sql = string.Format(
          "select * from Customers where City = '{0}'", city);
        Console.WriteLine(sql);
      }
    }

    static Func<NorthwindEntities, string, IEnumerable<Customer>> customersByCity = 
      CompiledQuery.Compile<NorthwindEntities, string, IEnumerable<Customer>>(
        (context, city) =>
          from c in context.Customers
          where c.City == city
          select c
      );

    static void CompiledQueries()
    {
      using (NorthwindEntities context = new NorthwindEntities())
      {
        var customers = customersByCity.Invoke(context, "London");
        foreach (Customer c in customers)
          Console.WriteLine(c.ContactName);
      }

      using (NorthwindEntities context = new NorthwindEntities())
      {
        var customers = customersByCity.Invoke(context, "London");
        customers = from c in customers
                    orderby c.ContactName
                    select c;
        foreach (Customer c in customers)
          Console.WriteLine(c.ContactName);
      }
    }

    static void ExtensionMethods()
    {
      using (NorthwindEntities context = new NorthwindEntities())
      {
        var customers = from c in context.Customers
                        where c.City == "London"
                        select c;
        foreach (Customer c in customers)
          Console.WriteLine(c.ContactName);
      }

      using (NorthwindEntities context = new NorthwindEntities())
      {
        var customers = context.Customers
                               .Where(c => c.City == "London")
                               .Select(c => c);
        foreach (Customer c in customers)
          Console.WriteLine(c.ContactName);
      }
    }

    static void EntityLifecycle()
    {
      using (NorthwindEntities context = new NorthwindEntities())
      {        
        Product p = new Product() { ProductName = "Mukuzani" };
        context.Products.AddObject(p);
        Console.WriteLine(context.ObjectStateManager.GetObjectStateEntry(p).State);
        context.SaveChanges();
        Console.WriteLine(context.ObjectStateManager.GetObjectStateEntry(p).State);
      }

      using (NorthwindEntities context = new NorthwindEntities())
      {
        Product product = context.Products.First(p => p.ProductName == "Mukuzani");
        Console.WriteLine(context.ObjectStateManager.GetObjectStateEntry(product).State);
        product.Category = context.Categories.First(c => c.CategoryName == "Beverages");
        Console.WriteLine(context.ObjectStateManager.GetObjectStateEntry(product).State);
        context.SaveChanges();
        Console.WriteLine(context.ObjectStateManager.GetObjectStateEntry(product).State);
      }

      using (NorthwindEntities context = new NorthwindEntities())
      {
        Product product = context.Products.First(p => p.ProductName == "Mukuzani");
        Console.WriteLine(context.ObjectStateManager.GetObjectStateEntry(product).State);
        context.Products.DeleteObject(product);
        Console.WriteLine(context.ObjectStateManager.GetObjectStateEntry(product).State);
        context.SaveChanges();
        Console.WriteLine(context.ObjectStateManager.GetObjectStateEntry(product).State);
      }
    }

    static void OptimisticLocking()
    {
      using (NorthwindEntities context1 = new NorthwindEntities())
      using (NorthwindEntities context2 = new NorthwindEntities())
      {
        Product product1 = context1.Products.First(p => p.ProductName == "Chai");
        Product product2 = context2.Products.First(p => p.ProductName == "Chai");

        product1.UnitsInStock = 50;
        product2.UnitsInStock = 20;

        context1.SaveChanges();
        context2.SaveChanges();
      }
    }

    static void OptimisticConcurrencyErrorHandling()
    {
      using (NorthwindEntities context = new NorthwindEntities())
      {
        Product product = context.Products.First(p => p.ProductName == "Chai");
        product.UnitsInStock = 50;
        try
        {
          context.SaveChanges();
        }
        catch (OptimisticConcurrencyException e)
        {
          context.Refresh(RefreshMode.StoreWins, e.StateEntries.Select(s => s.Entity));
        }
      }
    }

    static void Transactions()
    {
      using (TransactionScope transaction = new TransactionScope())
      using (NorthwindEntities context = new NorthwindEntities())
      {
        Category wines = context.Categories.FirstOrDefault(c => c.CategoryName == "Wines");
        if (wines == null)
        {
          wines = new Category() { CategoryName = "Wines" };
          context.Categories.AddObject(wines);
        }

        Product p = new Product() { ProductName = "Mukuzani", Category = wines };
        context.Products.AddObject(p);

        context.SaveChanges();
        transaction.Complete();
      }
    }

    static void PessimisticDeadLocking()
    {
      TransactionOptions to = new TransactionOptions();
      to.IsolationLevel = System.Transactions.IsolationLevel.RepeatableRead;
      TransactionScopeOption so = TransactionScopeOption.Required;
      using (TransactionScope transaction = new TransactionScope(so, to))
      using (NorthwindEntities context = new NorthwindEntities())
      {
        Product product = context.Products.First(p => p.ProductName == "Mukuzani");
        product.UnitsInStock -= 1;
        context.SaveChanges();
        transaction.Complete();
      }
    }

    static void StoredProcedures()
    {
      using (NorthwindEntities context = new NorthwindEntities())
      {
        IEnumerable<Product> products = context.GetProductsToReorder();
        foreach (Product p in products)
          Console.WriteLine(p.ProductName);
      }
    }

    static void Main(string[] args)
    {
      Inheritance.Examples.ReadBaseType();
      Console.ReadKey();
    }
  }
}
